@@ -284,6 +284,213 @@ def calculateLogGamma(uc, gradc, nu_t, tf_flag=False):
         gamma[gamma>1.0/constants.GAMMA_MIN] = 1.0/constants.GAMMA_MIN        
         return np.log(gamma)
     
+
+def cleanAnisotropy(b, test_inputs=None, b_default_rans=None, n_std=None, verbose=True):
+    """
+    This function is called to post-process the anisotropy b for realizability
+    
+    Arguments:
+    b -- numpy array of shape (n_useful, 3, 3) containing the originally predicted 
+         anisotropy tensor    
+    test_inputs -- numpy array of shape (n_useful, NUM_FEATURES) containing the test 
+                   point features that produced the anisotropy tensor we are dealing
+                   with. Optional, if None it is not used. 
+    b_default_rans -- numpy array containing the default anisotropy tensor in RANS.
+                      Shape is (num_points, 3, 3)
+    n_std -- float, number of standard deviations around the training mean that each
+             test feature is allowed to be. Optional, if None, default value is read from 
+             constants.py
+    verbose -- optional argument, boolean that says whether we print some extra 
+               information to the screen.
+    
+    Returns:
+    b -- numpy array of shape (n_useful, 3, 3) containing the post-processed
+         anisotropy tensor    
+    """
+    
+    
+    # Get default values if None is passed
+    if n_std is None:
+        n_std = constants.N_STD    
+    
+    print("Cleaning predicted anisotropy tensor... ", end="", flush=True)
+    tic = timeit.default_timer()        
+    
+    # (1) Here, make sure that there is no input more or less than n_std away
+    # from the mean. Since test_inputs was normalized, it should not be more
+    # than n_std or less than -n_std if we don't want test points too different
+    # from the training data. We are cleaning points that, by a very simple measure,
+    # denote extrapolation from the training set
+    mask_ext = np.zeros(b.shape[0], dtype=bool)
+    if test_inputs is not None:
+        mask_ext = (np.amax(test_inputs,axis=1) > n_std) + \
+                   (np.amin(test_inputs,axis=1) < -n_std)        
+    num_ext = np.sum(mask_ext) # total number of entries affected by this step
+    if num_ext > 0: b[mask_ext, :, :] = b_default_rans[mask_ext, :, :]
+    #--------------------------------------------------------------------   
+    
+    # (2) Here, make sure all matrix entries are realizable.
+    b, n_real, _, = makeRealizableFast(b)    
+    #--------------------------------------------------------------------                 
+    
+    # Make sure no tensor entry disobeys the criterium in Ling et al. JFM 2016
+    mask = np.zeros(b.shape[0], dtype=bool)
+    for i in range(3):
+        for j in range(3):
+            if i == j:
+                mask = mask + (b[:,i,j] > 2.0/3)
+                mask = mask + (b[:,i,j] < -1.0/3)
+            else:
+                mask = mask + (b[:,i,j] > 0.5)
+                mask = mask + (b[:,i,j] < -0.5)
+    eig_all, _ = np.linalg.eigh(b)
+    mask = mask + (eig_all[:, 2] < (3.0*np.abs(eig_all[:, 1]) - eig_all[:, 1])/2.0)
+    mask = mask + (eig_all[:, 2] > 1.0/3 - eig_all[:, 1])
+    n_fail = np.sum(mask)    
+    
+    # Print information        
+    toc = timeit.default_timer()
+    print("Done! It took {:.1f}s".format(toc-tic), flush=True)
+    if verbose:
+        if test_inputs is not None:
+            print("{} points ({:.2f}% of total) were cleaned due to outlier inputs."\
+                  .format(num_ext, 100.0*num_ext/b.shape[0]), flush=True)               
+        print("{} points ({:.2f}% of total) were cleaned due to non-realizable tensors."\
+              .format(n_real, 100.0*n_real/b.shape[0]), flush=True)
+        print("In cleaned anisotropy tensor: {} violations detected".format(n_fail))
+    
+    return b
+
+
+def makeRealizableFast(b):
+    """
+    This function takes in Reynolds stress anisotropy tensor and makes it realizable
+    
+    This function is based on code written by Dr. Andrew Banko. It checks that each
+    point represents a realizable turbulence state by verifying that it lies within 
+    the barycentric map (c.f. Banerjee et al., Journal of Turbulence, 2017 and
+    Emory and Iaccarino, CTR Research Briefs, 2014). If a point lies outside the 
+    triangle, it is moved to the nearest point on the boundary of the triangle. 
+    
+    Arguments:  
+    b -- a numpy array of shape (num_points, 3, 3) containing the Reynolds
+         stress anisotropy components
+    
+    Returns:
+    b -- corrected anisotropy tensor, shape (num_points, 3, 3)
+    n_fails -- number of points failing the realizability test
+    ind_fail -- numpy array of shape (num_points,) containing 1 if the original data
+                was not realizable, and 0 if it was realizable
+    """
+
+    # Functions to transform between eigenvalues, barycentric weights, and barycentric coordinates
+    def eval_to_C(evalues):
+        C1 = evalues[0] - evalues[1]
+        C2 = 2.*(evalues[1] - evalues[2])
+        C3 = 3.*evalues[2] + 1.
+        C3 = 1. - C1 - C2 # Uncomment to verify sum(Ci) = 1
+        return C1, C2, C3
+
+    def C_to_x(C1, C2, C3):
+        x = np.zeros(2)
+        x[0] = C1 + 0.5*C3
+        x[1] = C3*np.sqrt(3)/2.
+        return x
+
+    def x_to_C(x):
+        C3 = x[1]*2./np.sqrt(3)
+        C1 = x[0] - 0.5*C3
+        C2 = 1. - C1 - C3
+        return C1, C2, C3
+
+    def C_to_eval(C1, C2, C3):
+        evalues = np.zeros(3)
+        evalues[0] = C1 + 0.5*C2 + C3/3. - 1./3.
+        evalues[1] = 0.5*C2 + C3/3. - 1./3.
+        evalues[2] = C3/3. - 1./3.
+        return evalues            
+
+    # Get number of points, initialize number of realizability violations
+    num_points = b.shape[0]
+    n_fails = 0
+    ind_fail = np.zeros((num_points))
+
+    # Unit vectors defining edges opposite vertices 1, 2, 3
+    e1 = np.array([0.5, np.sqrt(3)/2.])
+    e1 = e1 / np.linalg.norm(e1)
+    e2 = e1
+    e2[0] = -1.*e2[0]
+    e3 = np.array([1., 0.])
+    e1perp = np.array([e1[1], -1.*e1[0]])
+    e2perp = np.array([-1.*e2[1], e2[0]])
+    e3perp = np.array([e3[1], e3[0]])
+    eperp_mat = np.transpose(np.array([e1perp, e2perp, e3perp]))
+    pert = 1.e-8
+    # Array of vertices 1, 2, 3: i.e. v1 = v[:, 0]
+    v = np.array([[1., 0., 0.5], [0., 0., np.sqrt(3)/2.]]) 
+    
+    # Verify that b is symmetric with zero trace
+    b = 0.5 * (b + np.transpose(b, (0, 2, 1)))
+    trace_b = np.trace(b, axis1=1, axis2=2)
+    for i in range(3):
+        b[:, i, i] = b[:, i, i] - 1./3. * trace_b
+
+    # Loop over points to check realizability and project to barycentric map if necessary
+    for i in range(num_points):
+        # Initialize point's realizability indicator
+        fail_ID = 0
+        
+        # Compute eigenvalues and eigenvectors
+        this_b = b[i, :, :]
+        evalues, evectors = np.linalg.eig(this_b)
+        sort = np.argsort(evalues)
+        sort = np.flip(sort,axis=0)
+        evalues = evalues[sort]
+        evectors = evectors[:, sort]
+        
+        # Check realizability, correct if necessary
+        # Correction procedure: for each vertex, determine if outside triangle
+        # opposite edge - if yes, project to nearest point on that edge.
+        # At end, if still outside triangle, go to nearest vertex.
+        C1, C2, C3 = eval_to_C(evalues)
+        x = C_to_x(C1, C2, C3)
+        if C1 < 0.0:
+            fail_ID = 1
+            x = np.dot(x, e1)*e1 + pert*e1perp 
+            C1, C2, C3 = x_to_C(x)
+        if C2 < 0.0:
+            fail_ID = 1
+            x = np.dot(x - e3, e2)*e2 + e3 + pert*e2perp
+            C1, C2, C3 = x_to_C(x)
+        if C3 < 0.0:
+            fail_ID = 1
+            x = np.dot(x, e3)*e3 + pert*e3perp
+            C1, C2, C3 = x_to_C(x)
+        if (C1 < 0.0) or (C2 < 0.0) or (C3 < 0.0):
+            delta_v = np.zeros(3)
+            for j in range(3):
+                v_temp = v[:, j]
+                delta_v[j] = np.linalg.norm(x - v_temp)
+            ind_min = np.argmin(delta_v)
+            x = v[:, ind_min] + pert*(eperp_mat[:, np.mod(ind_min-1, 3)] + eperp_mat[:, np.mod(ind_min+1, 3)])
+
+        # Compute corrected anisotropy tensor
+        if fail_ID == 1:
+            evalues = C_to_eval(C1, C2, C3)
+            this_b = np.dot(np.dot(evectors, np.diag(evalues)), np.linalg.inv(evectors))
+            b[i, :, :] = this_b
+            n_fails += 1
+
+        ind_fail[i] = fail_ID
+
+    # Verify that b is symmetric with zero trace
+    b = 0.5 * (b + np.transpose(b, (0, 2, 1)))
+    trace_b = np.trace(b, axis1=1, axis2=2)
+    for i in range(3):
+        b[:, i, i] = b[:, i, i] - 1./3. * trace_b   
+
+    return b, n_fails, ind_fail
+
     
 def suppressWarnings():
     """
@@ -293,5 +500,4 @@ def suppressWarnings():
     if type(tf.contrib) != type(tf): tf.contrib._warning = None
     os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
     tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
-    os.environ['KMP_WARNINGS'] = 'off'
-    
\ No newline at end of file
+    os.environ['KMP_WARNINGS'] = 'off'    
\ No newline at end of file
